import datetime
import json
import os
import random

from sideseeing_tools import (
    constants, 
    exceptions,
    media,
    utils,
)


class SideSeeingDS:
    def __init__(
            self, 
            root_dir,
            subdir_data='.',
            name='MyDataset', 
            generate_metadata=False,
            extract_media=False,
        ):
        print('INFO. Loading data.')
        self.name = name
        
        if not os.path.isdir(root_dir):
            raise exceptions.RootDirIsNotADirectoryError()

        if not os.path.exists(root_dir):
            raise exceptions.RootDirDoesNotExistError()

        self.root_dir = root_dir if root_dir.endswith(os.path.sep) else f'{root_dir}{os.path.sep}'
        self.data_dir = os.path.join(self.root_dir, subdir_data)
        self.setup(extract_media)

        if generate_metadata:
            self.metadata(generate_metadata)
        print('INFO. Done.')

    def setup(self, extract_media):
        self.instances = {}
        for root, _, files in os.walk(self.data_dir):
            for f in files:
                if f not in constants.SUPPORTED_FILES:
                    print(f'WARNING. {os.path.join(root, f)} has been ignored.')
                    continue

                ssf = SideSeeingFile(self.data_dir, os.path.join(root, f))
                if ssf.is_valid:
                    if ssf.name not in self.instances:
                        self.instances[ssf.name] = SideSeeingInstance(
                                ssf.name,
                                ssf.path,
                            )
                    self.instances[ssf.name].add_file(ssf)

        for key in self.instances.keys():
            self.instances[key].setup(extract_media)

        self.populate_sensors()

    def populate_sensors(self):
        self.sensors = {
            'sensors1': {},
            'sensors3': {},
            'sensors6': {},
        }

        for instance in self.iterator:
            for n_axis in self.sensors.keys():
                if hasattr(instance, n_axis):
                    for name in getattr(instance, n_axis).keys():
                        if name not in self.sensors[n_axis]:
                            self.sensors[n_axis][name] = set()
                        self.sensors[n_axis][name].add(instance.name)

    @property
    def instance(self):
        return self.instances[random.choice(list(self.instances.keys()))]

    @property
    def size(self):
        return len(self.instances)

    @property
    def iterator(self):
        for k in sorted(self.instances.keys()):
            yield self.instances[k]

    def metadata(self, save=False):
        if self.size == 0:
            print(f'ERROR. Dataset is empty.')
            return

        path = os.path.join(f'{self.root_dir}', 'metadata.csv')

        if os.path.exists(path) and not save:
            df = utils.load_csv_data_with_pandas(path)

        if not os.path.exists(path) or save:
            df = utils.generate_metadata(self.iterator, constants.DATETIME_UTC_FORMAT)
            utils.save_csv_data_with_pandas(df, path)

        return df

    def __str__(self):
        return f'SSDS[name: {self.name}, instances: {self.size}]'

    def __repr__(self):
        return self.__str__()


class SideSeeingFile:
    def __init__(self, data_dir, path):
        self.data_dir = data_dir
        self.file_path = path
        self.setup()

    def setup(self):
        self.file_name = os.path.basename(self.file_path)
        self.file_type = self.discover_file_type()
        self.path = os.path.dirname(self.file_path)
        self.name = self.gen_instance_name(self.file_path, self.data_dir)
        self.is_valid = False if self.file_type == 'unknown' else True

    def gen_instance_name(self, file_path, data_dir):
        els = f'{os.path.dirname(file_path)}'.replace(data_dir, '').split(os.path.sep)
        return '#'.join(els[1:])

    def discover_file_type(self):
        if self.file_name.endswith('metadata.json'):
            return 'metadata'
        elif (self.file_name.endswith('labels.txt') or self.file_name.endswith('labels.csv')):
            return 'label'
        elif self.file_name.endswith('consumption.csv'):
            return 'consumption'
        elif self.file_name.endswith('gps.csv'):
            return 'gps'
        elif self.file_name.endswith('sensors.three.csv'):
            return 'sensors3'
        elif self.file_name.endswith('sensors.three.uncalibrated.csv'):
            return 'sensors6'
        elif self.file_name.endswith('sensors.one.csv'):
            return 'sensors1'
        elif self.file_name.endswith('.mp4'):
            return 'video'
        elif self.file_name.endswith('video.wav'):
            return 'audio'
        elif self.file_name.endswith('video.gif'):
            return 'gif'
        else:
            print(f'WARNING. Unknown file type detected: {self.file_path}.')
            return 'unknown'

    def __str__(self):
        return f'SSFile[file_path: {self.file_path}]'

    def __repr__(self):
        return self.__str__()


class SideSeeingInstance:
    def __init__(self, name, path):
        self.name = name
        self.path = path
        self.files = {}

    def add_file(self, ssf: SideSeeingFile):
        self.files[ssf.file_type] = ssf

    def __str__(self):
        return f'SSInstance[name: {self.name}]'

    def __repr__(self):
        return self.__str__()

    def print_metadata(self):
        print(f"----\nName: {self.name}")
        print(
            f"Path: {self.path}",
            f"Manufacturer: {self.metadata.get('device', {}).get('manufacturer', '').title()}",
            f"Model: {self.metadata.get('device', {}).get('model', '')}",
            f"Android version: {self.metadata.get('device', {}).get('androidVersion', '')}",
            f"Start video time: {self.metadata.get('time', {}).get('videoStartDateTime', '')}",
            f"Stop video time: {self.metadata.get('time', {}).get('videoStopDateTime', '')}",
            sep='\n'
        )

    def setup(self, extract_media=False):
        try:
            with open(self.files['metadata'].file_path) as json_file:
                self.metadata = json.load(json_file)
        except KeyError:
            raise exceptions.MetadataFileDoesNotExistError(f'ERROR. Metadata file is missing for {self.name}')

        if self.metadata is None:
            raise exceptions.InvalidMetadataFileError(f'ERROR. Metadata file is is invalid for {self.name}')

        self.video_start_time = datetime.datetime.strptime(self.metadata['time']['videoStartDateTime'], constants.DATETIME_UTC_FORMAT)
        self.video_stop_time = datetime.datetime.strptime(self.metadata['time']['videoStopDateTime'], constants.DATETIME_UTC_FORMAT)
        self.video_total_time = (self.video_stop_time - self.video_stop_time).total_seconds()

        for k, v in self.files.items():
            if k == 'consumption':
                self.consumption = utils.load_csv_data(v.file_path, fieldnames=constants.CONSUMPTION_FILE_FIELDNAMES)

            if k == 'gps':
                self.geolocation_points = utils.preprocess_gps(
                    utils.load_csv_data(v.file_path, fieldnames=constants.GPS_FILE_FIELDNAMES)
                )
                self.geolocation_center = self.geolocation_points.mean(axis=0)

            if k == 'sensors3':
                self.sensors3 = utils.preprocess_sensors(
                    utils.load_csv_data(v.file_path, fieldnames=constants.THREE_AXES_SENSORS_FILE_FIELDNAMES),
                    3,
                    constants.DATETIME_UTC_FORMAT,
                    self.video_start_time,
                    self.video_stop_time
                )

            if k == 'sensors6':
                self.sensors6 = utils.preprocess_sensors(
                    utils.load_csv_data(v.file_path, fieldnames=constants.THREE_AXES_UNCALIBRATED_SENSORS_FILE_FIELDNAMES),
                    6,
                    constants.DATETIME_UTC_FORMAT,
                    self.video_start_time,
                    self.video_stop_time
                )

            if k == 'sensors1':
                self.sensors1 = utils.preprocess_sensors(
                    utils.load_csv_data(v.file_path, fieldnames=constants.ONE_AXIS_SENSORS_FILE_FIELDNAMES),
                    1,
                    constants.DATETIME_UTC_FORMAT,
                    self.video_start_time,
                    self.video_stop_time
                )

            if k == 'label':
                self.label = utils.load_csv_data(v.file_path, fieldnames=constants.LABELS_FILE_FIELDNAMES)

            if k == 'video':
                self.video = v.file_path
                if extract_media:
                    self.audio = media.extract_audio(v.file_path, v.file_path.replace('.mp4', '.wav'))
                    self.gif = media.extract_gif(v.file_path, v.file_path.replace('.mp4', '.gif'))

    def extract_snippet(self, start_time, end_time, output_dir):
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        sensor_dicts = []
        if hasattr(self, 'sensors1'):
            sensor_dicts.append(self.sensors1)

        if hasattr(self, 'sensors3'):
            sensor_dicts.append(self.sensors3)

        if hasattr(self, 'sensors6'):
            sensor_dicts.append(self.sensors6)

        for sensor_dict in sensor_dicts:
            for s, sensor in sensor_dict.items():
                utils.extract_sensor_snippet(
                    sensor,
                    start_time,
                    end_time,
                    output_path=os.path.join(output_dir, f'{s.lower()}_{start_time}_{end_time}.csv')
                )

        media.extract_video_snippet(
            self.video, 
            start_time, 
            end_time, 
            os.path.join(output_dir, f'video_{start_time}_{end_time}.mp4'),
        )



import cv2
import folium
import librosa
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from matplotlib import animation

from sideseeing_tools import constants, media, utils
from sideseeing_tools import sideseeing as sst


class SideSeeingPlotter:
    def __init__(self, dataset: sst.SideSeeingDS, path_taxonomy: str=None, google_api_key: str=None):
        self.dataset = dataset
        if path_taxonomy:
            self.categories_and_tags = utils.load_csv_data(path_taxonomy, fieldnames=constants.LABELS_FILE_FIELDNAMES)
        else:
            self.categories_and_tags = []
        self.google_api_key = google_api_key

    def plot_instance_audio(self, instance: sst.SideSeeingInstance):
        '''
        Plots the waveform and the Mel spectrogram for the specified audio file.
        '''
        if not hasattr(instance, 'audio'):
            print(f'Extracting audio data')
            instance.audio = media.extract_audio(
                instance.video, 
                instance.video.replace('.mp4', '.wav')
            )

        y, sr = librosa.load(instance.audio)
        M = librosa.feature.melspectrogram(y=y, sr=sr)
        M_db = librosa.power_to_db(M, ref=np.max)

        fig, ax = plt.subplots(nrows=2, sharex=True)
        librosa.display.waveshow(y, sr=sr, ax=ax[0], color="blue")
        ax[0].set(title='Waveform')
        ax[0].label_outer()

        img1 = librosa.display.specshow(M_db, y_axis='mel', x_axis='time', ax=ax[1])
        ax[1].set(title='Mel spectrogram')
        fig.colorbar(img1, ax=ax[1], format="%+2.f dB", location='bottom')
        plt.tight_layout()
        plt.show()

    def plot_instance_map(self, instance: sst.SideSeeingInstance, titles='OpenStreetMap', zoom_start=14):        
        points = instance.geolocation_points
        center = instance.geolocation_center

        if len(points) > 0:
            map = folium.Map(location=center, titles=titles, zoom_start=zoom_start)
            counter = 1

            for lat, lon in points:
                if counter == 1:
                    color = 'red'
                elif counter == len(points):
                    color = 'green'
                else:
                    color = 'blue'

                folium.Marker([lat, lon], icon=folium.Icon(color=color)).add_to(map)
                counter += 1

            folium.LayerControl().add_to(map)
            return map
        else:
            print('WARNING. GPS data is missing.')

    def plot_instance_video_frames(self, instance: sst.SideSeeingInstance):
        '''
        Plots a sample of frames for the specified video.
        '''
        cap = cv2.VideoCapture(instance.video)
        fps = cap.get(cv2.CAP_PROP_FPS)
        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        frame_indices = [0] + [int(i * total_frames / 10) for i in range(1, 10)]

        rows = 2
        cols = 5
        fig, axes = plt.subplots(rows, cols, figsize=(15, 6))

        for i, idx in enumerate(frame_indices):
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

            time_in_seconds = idx / fps

            ax = axes[i // cols, i % cols]
            ax.imshow(frame_rgb)
            ax.set_title(f'Frame {idx} - {time_in_seconds:.1f} s')

            ax.set_xticks([])
            ax.set_yticks([])

        plt.tight_layout()
        plt.show()
        cap.release()

    def plot_instance_video_frames_at_times(self, instance: sst.SideSeeingInstance, times: list, show_frame_number=True):
        '''
        Plots frames for the specified video at the given time points.
        '''
        cap = cv2.VideoCapture(instance.video)
        fps = cap.get(cv2.CAP_PROP_FPS)
        num_frames = len(times)

        max_cols = 5

        if num_frames <= max_cols:
            rows = 1
            cols = num_frames
            figsize = (15, 3)
        else:
            rows = int(num_frames / max_cols) + 1
            cols = 5
            figsize = (15, 2 * rows)

        fig, axes = plt.subplots(rows, cols, figsize=figsize)

        rc = 0
        cc = 0
        for i, time_in_seconds in enumerate(times):
            frame_idx = int(time_in_seconds * fps)

            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_idx)

            ret, frame = cap.read()

            frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

            if rows > 1:
                ax = axes[rc][cc]
            else:
                ax = axes[cc]

            ax.imshow(frame_rgb)
            if show_frame_number:
                ax.set_title(f'{time_in_seconds:.1f} s' )

            ax.set_xticks([])
            ax.set_yticks([])

            if cc < max_cols - 1:
                cc += 1
            else:
                cc = 0
                rc += 1

        [fig.delaxes(ax) for ax in axes.flatten() if not ax.has_data()]
        plt.tight_layout()
        plt.show()
        cap.release()

    def plot_instance_sensors3_and_audio(self, instance: sst.SideSeeingInstance, main_title='', x_ti=0, x_tf=None):
        '''
        Creates a plot for the sensors and audio data.
        It is useful for plotting data related with an instance/sample.
        '''
        num_sensor_subplots = len(instance.sensors3.keys())
        num_audio_subplots = 2
        total_subplots = num_sensor_subplots + num_audio_subplots

        fig, axis = plt.subplots(total_subplots, 2, figsize=(15, 15), width_ratios=[3, 1], sharex=False)
        fig.suptitle(main_title, fontsize=15)

        for ind, sensor_name in enumerate(sorted(instance.sensors3.keys())):
            sensor_data = instance.sensors3[sensor_name]

            if not x_tf:
                x_tf = sensor_data['Time (s)'].iloc[-1]

            axis[ind, 0].set_ylabel(sensor_name)

            axis[ind, 0].plot(sensor_data['Time (s)'], sensor_data['x'], label='x', linewidth=0.75)
            axis[ind, 0].plot(sensor_data['Time (s)'], sensor_data['y'], label='y', linewidth=0.75)
            axis[ind, 0].plot(sensor_data['Time (s)'], sensor_data['z'], label='z', linewidth=0.75)
            axis[ind, 0].set_xlabel('Time (s)')
            axis[ind, 0].set_xlim((x_ti, x_tf))

            axis[ind, 1].hist(sensor_data['x'], label='x', alpha=0.4)
            axis[ind, 1].hist(sensor_data['y'], label='y', alpha=0.4)
            axis[ind, 1].hist(sensor_data['z'], label='z', alpha=0.4)
            axis[ind, 1].set_ylabel('Frequency')

            axis[ind, 0].legend()
            axis[ind, 1].legend()

            if ind == num_sensor_subplots - 1:
                if not hasattr(instance, 'audio'):
                    print(f'Extracting audio data')
                    instance.audio = media.extract_audio(
                        instance.video, 
                        instance.video.replace('.mp4', '.wav')
                    )

                y, sr = librosa.load(instance.audio)
                M = librosa.feature.melspectrogram(y=y, sr=sr)
                M_db = librosa.power_to_db(M, ref=np.max)

                librosa.display.waveshow(y, sr=sr, ax=axis[ind + 1, 0], color="blue")
                axis[ind + 1, 0].set_xlim((x_ti, x_tf))
                axis[ind + 1, 0].set_ylabel('Waveform')
                axis[ind + 1, 0].set_xlabel('Time (s)')

                img = librosa.display.specshow(
                        M_db,
                        y_axis='mel',
                        x_axis='time',
                        ax=axis[ind + 2, 0]
                )
                axis[ind + 2, 0].set_xlim((x_ti, x_tf))
                axis[ind + 2, 0].set_ylabel('Mel spectrogram (Hz)')
                axis[ind + 2, 0].set_xlabel('Time (s)')

                fig.colorbar(
                        img,
                        ax=axis[ind + 2, 1],
                        format="%+2.f dB",
                        location='left'
                )

                axis[ind + 1, 1].axis('off')
                axis[ind + 2, 1].axis('off')

        plt.tight_layout()
        plt.show()

    def plot_dataset_tags(self):
        if len(self.categories_and_tags) == 0:
            print('WARNING. There are no categories and tags loaded. Check the taxonomy file.')
            return

        labels = {}
        for l in self.categories_and_tags:
            cat, tag = l['category'], l['tag']
            k = f'{cat},{tag}'.replace(' ', '_')
            labels[k] = 0

        for instance in self.dataset.iterator:
            if not hasattr(instance, 'label'):
                print(f"WARNING. {instance.name} isn't labeled.")
            else:
                for l in instance.label:
                    cat, tag = l['category'], l['tag']
                    k = f'{cat},{tag}'.replace(' ', '_')
                    if k not in labels:
                        labels[k] = 0
                    labels[k] += 1

        df = pd.DataFrame.from_dict(data=labels, orient='index', columns=['Frequency'])
        df = df.fillna(0)
        df = df.astype(int)
        df = df.sort_values('Frequency', ascending=False)

        return df.style.applymap(func=lambda x: 'color: red' if x < 1 else 'color: black')
    
    def plot_dataset_tags_matrix(self, primary_category, secondary_category):
        if len(self.categories_and_tags) == 0:
            print('WARNING. There are no categories and tags loaded. Check the taxonomy file.')
            return
        
        if primary_category is None or secondary_category is None:
            print(f'ERROR. You must pass a category name.')
            print(f'Here are the possibilities: {self.categories_and_tags.keys()}.')
            return

        matrix = {}

        for ct in self.categories_and_tags:
            c = ct['category']
            t = ct['tag']

            if c == primary_category:
                matrix[t] = {}

        for st in matrix:
            for ct in self.categories_and_tags:
                c = ct['category']
                t = ct['tag']

                if c == secondary_category:
                    matrix[st][t] = 0

        for instance in self.dataset.iterator:
            if not hasattr(instance, 'label'):
                print(f"WARNING. {instance.name} isn't labeled.")
            else:
                instance_st = set()
                for l in instance.label:
                    cat, tag = l['category'], l['tag']
                    if cat == primary_category:
                        instance_st.add(tag)

                for l in instance.label:
                    cat, tag = l['category'], l['tag']
                    if cat == secondary_category:
                        for s in instance_st:
                            matrix[s][tag] += 1

        df = pd.DataFrame(data=matrix)
        df = df.fillna(0)
        df = df.astype(int)

        return df.style.applymap(func=lambda x: 'color: red' if x < 1 else 'color: black')
    
    def plot_dataset_map(self, titles='OpenStreetMap', zoom_start=4):
        points = np.vstack([i.geolocation_points for i in self.dataset.iterator])
        center = points.mean(axis=0)

        ds_map = folium.Map(location=center, titles=titles, zoom_start=zoom_start)

        for lat, lon in points:
            folium.Marker([lat, lon]).add_to(ds_map)

        folium.LayerControl().add_to(ds_map)
        return ds_map

    def plot_dataset_cities(self):
        points = np.vstack([i.geolocation_center for i in self.dataset.iterator])

        geo = {}
        for lat, long in points:
            location = utils.inverse_geocode(lat, long, self.google_api_key)
            key = f"{location['country']},{location['city']}"

            if key not in geo:
                geo[key] = {'Frequency': 0}

            geo[key]['Frequency'] += 1

        df = pd.DataFrame.from_dict(data=geo, orient='index', columns=['Frequency'])
        df = df.fillna(0)
        df = df.sort_values('Frequency', ascending=False)

        return df.style.applymap(func=lambda x: 'color: red' if x < 1 else 'color: black')

    def generate_video_for_sensor(self, data, time_column, axis_columns, output_path, start_second=None, end_second=None, xlim=None, ylim=None, use_dynamic_xaxis=False, dpi=90, figsize = (25, 4), linewidth=5):
        if start_second and end_second:
            data = data[(data['Time (s)'] >= start_second) & (data['Time (s)'] <= end_second)]
    
        data['Time (s)'] = data['Time (s)'] - data['Time (s)'].iloc[0]
        data = data.reset_index(drop=True)

        plt.ion()

        mpl.rcParams['lines.linewidth'] = linewidth
        mpl.rcParams['font.family'] = "monospace"

        fig, ax = plt.subplots(1, 1, figsize = figsize)

        time_min = np.min(data[time_column])
        time_max = np.max(data[time_column])
        axis_min = np.min(np.min(data[axis_columns]))
        axis_max = np.max(np.max(data[axis_columns]))
        axis_dt = int(abs((axis_max - axis_min)) * 0.1) or 1

        fps = int(len(data['x'])/time_max)

        def animate(i):
            ax.cla()

            ax.xaxis.set_visible(False)
            ax.yaxis.set_visible(False)
            ax.set_facecolor((0, 0, 0, 1))
            ax.set_frame_on(False)
            fig.patch.set_alpha(1)

            if use_dynamic_xaxis:
                x_start = max(0, i - 100)
                x_end = i
                ax.set_xlim([data[time_column][x_start], data[time_column][x_end]])
            else:
                ax.set_xlim([time_min, time_max])

            if xlim:
                ax.set_xlim(xlim)
            else:
                ax.set_xlim([time_min, time_max])

            if ylim:
                ax.set_ylim(ylim)
            else:
                ax.set_ylim([axis_min - axis_dt, axis_max + axis_dt])

            for col in axis_columns:
                ax.plot(data[time_column][:i], data[col][:i])

        anim = animation.FuncAnimation(fig, animate, frames=len(data['x']), interval=200, blit=False, repeat=False)
        plt.subplots_adjust(left=0.01, right=1, top=1, bottom=0.01)
        anim.save(
            output_path,
            codec='libx264rgb',
            dpi=dpi,
            fps=fps,
            writer='ffmpeg',
            savefig_kwargs={
            "transparent": True,
            "facecolor": "Black"
            }
        )

    def plot_dataset_frames_at_times(self, times, show_instance_name=True, show_frame_number=True):
        for instance in self.dataset.iterator:
            if show_instance_name:
                print(instance.name)
            self.plot_instance_video_frames_at_times(instance, times=times, show_frame_number=show_frame_number)

    def plot_sensor(self, data, time_column, axis_columns, xlim=None, ylim=None, title=None, linewidth=0.75, figsize=(20, 4)):
        fig, ax = plt.subplots(nrows=1, figsize=figsize)

        for ind, a in enumerate(axis_columns):
            ax.plot(data[time_column], data[a], label=axis_columns[ind], linewidth=linewidth)

        ax.set_title(title)

        if xlim:
            ax.set_xlim(xlim)

        if ylim:
            ax.set_ylim(ylim)

        ax.legend()
        fig.show()

    def plot_sensors(self, data, time_column, axis_columns, xlim=None, ylim=None, linewidth=0.75, sharex=True, sharey=True):
        if len(data) == 0:
            return

        _, axis = plt.subplots(len(data), 1, figsize=(15, int(len(data) * 4)), sharex=sharex, sharey=sharey)

        for ind_plot, item in enumerate(data):
            instance_name = item['instance_name']
            sensor_data = item['sensor_data']

            for ind, column in enumerate(axis_columns):
                axis[ind_plot].plot(sensor_data[time_column], sensor_data[column], label=axis_columns[ind], linewidth=linewidth)
                axis[ind_plot].set_title(instance_name)
                axis[ind_plot].legend()
    
            if xlim:
                axis[ind_plot].set_xlim(xlim)
            if ylim:
                axis[ind_plot].set_ylim(ylim)

        plt.tight_layout()
        plt.show()



class RootDirIsNotADirectoryError(Exception):
  ...

class RootDirDoesNotExistError(Exception):
  ...

class MetadataFileDoesNotExistError(Exception):
  ...

class InvalidMetadataFileError(Exception):
  ...


DATETIME_UTC_FORMAT ="%Y-%m-%dT%H:%M:%S.%fZ"

CONSUMPTION_FILE_FIELDNAMES = [
  'datetime_utc',
  'battery_microamperes',
]

GPS_FILE_FIELDNAMES = [
  'datetime_utc',
  'gps_interval',
  'accuracy',
  'latitude',
  'longitude',
]

LABELS_FILE_FIELDNAMES = [
  'category',
  'tag',
]

ONE_AXIS_SENSORS_FILE_FIELDNAMES = [
  'timestamp_nano',
  'datetime_utc',
  'name',
  'axis_x',
  'accuracy',
]

THREE_AXES_SENSORS_FILE_FIELDNAMES = [
  'timestamp_nano',
  'datetime_utc',
  'name',
  'axis_x',
  'axis_y',
  'axis_z',
  'accuracy',
]

THREE_AXES_UNCALIBRATED_SENSORS_FILE_FIELDNAMES = [
  'timestamp_nano',
  'datetime_utc',
  'name',
  'axis_x',
  'axis_y',
  'axis_z',
  'delta_x',
  'delta_y',
  'delta_z',
  'accuracy',
]

SUPPORTED_FILES = [
  'consumption.csv',
  'gps.csv',
  'metadata.json',
  'sensors.three.csv',
  'video.gif',
  'video.wav',
  'labels.txt',
  'labels.csv',
  'sensors.one.csv',
  'sensors.three.uncalibrated.csv',
  'video.mp4',
]



import os

import imageio
import cv2

from moviepy.editor import VideoFileClip


def extract_audio(source_path: str, target_path: str, sample_rate=44100, channels=2, codec='pcm_s16le', overwrite=False):
    '''
    Extracts audio in WAV format from a video file.
    '''
    target_dir = os.path.dirname(target_path)

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    mp4_name = os.path.basename(source_path)
    wav_path = os.path.join(target_dir, mp4_name.replace('.mp4', '.wav'))

    if not os.path.exists(wav_path) or overwrite:
        print(f'INFO. Extracting WAV from {source_path} to {wav_path}.')
        clip = VideoFileClip(source_path)
        clip.audio.write_audiofile(wav_path, fps=sample_rate, nbytes=2, codec=codec, ffmpeg_params=["-ac", str(channels)])
        clip.close()

    return wav_path


def extract_gif(source_path: str, target_path: str, target_width=300, target_fps=5):
    '''
    Extracts a tiny video in GIF format from a video file.
    '''
    target_dir = os.path.dirname(target_path)

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    mp4_name = os.path.basename(source_path)
    gif_path = os.path.join(target_dir, mp4_name.replace('.mp4', '.gif'))

    if not os.path.exists(gif_path):
        print(f'INFO. Extracting GIF from {source_path} to {gif_path}.')

        cap = cv2.VideoCapture(source_path)
        original_fps = cap.get(cv2.CAP_PROP_FPS)

        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        target_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT) * (target_width / cap.get(cv2.CAP_PROP_FRAME_WIDTH)))
        target_resolution = (target_width, target_height)

        gif_frames_list = []
        frame_positions = [i for i in range(0, total_frames, int(original_fps) * 4)]

        for i in frame_positions:
            cap.set(cv2.CAP_PROP_POS_FRAMES, i)
            ret, frame = cap.read()

            if ret:
                if frame.shape[2] == 3:
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

                resized_frame = cv2.resize(frame, target_resolution)
                gif_frames_list.append(resized_frame)
            else:
                break

        imageio.mimsave(gif_path, gif_frames_list, fps=target_fps, subrectangles=True)

        cap.release()

    return gif_path


def extract_video_snippet(source_path, start_second, end_second, output_path):
    try:
        clip = VideoFileClip(source_path)
        snippet = clip.subclip(start_second, end_second)
        snippet.write_videofile(output_path)

    except FileNotFoundError:
        print(f'ERROR. Source file not found. Check the file path {source_path}.')

    except Exception as e:
        print(f'Error occurred: {str(e)}')

    finally:
        if 'clip' in locals():
            clip.close()



import csv
import datetime
import requests

import cv2
import numpy as np
import pandas as pd
import reverse_geocode


def load_csv_data(path: str, fieldnames: list, delimiter=','):
    '''
    Reads a CSV file using csv.DictReader and a predefined list of fields.
    '''
    data = []

    with open(path) as fin:
        reader = csv.reader(fin, delimiter=delimiter)
        first_row = next(reader)
        
        if [field.strip().lower() for field in first_row] == [field.strip().lower() for field in fieldnames]:
            fin.seek(0)
            reader = csv.DictReader(fin, delimiter=delimiter)
        else:
            reader = csv.DictReader(fin, fieldnames=fieldnames, delimiter=delimiter)

        for row in reader:
            new_row = {k.strip().lower(): v.strip().lower() for k, v in row.items()}
            data.append(new_row)

    return data


def load_csv_data_with_pandas(path: str):
    return pd.read_csv(path)


def save_csv_data_with_pandas(data: pd.DataFrame, path: str, index=False):
    data.to_csv(path, index=index)


def generate_metadata(iterator, datetime_format: str):
    items = []

    for i in iterator:
        cap = cv2.VideoCapture(i.video)
        v_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        v_fps = cap.get(cv2.CAP_PROP_FPS) or 30.0
        v_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        v_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        cap.release()

        item = {}

        item['name'] = i.name
        
        if hasattr(i, 'geolocation_center'):
          item['geolocation_center'] = f'{i.geolocation_center[0]}, {i.geolocation_center[1]}'
        else:
           item['geolocation_center'] = ''

        item['video_start_time'] = datetime.datetime.strptime(i.metadata.get('time', {}).get('videoStartDateTime', ''), datetime_format)
        item['video_end_time'] = datetime.datetime.strptime(i.metadata.get('time', {}).get('videoStopDateTime', ''), datetime_format)
        item['video_duration'] = round(v_frames / v_fps, 2)
        item['video_frames'] = v_frames
        item['video_fps'] = v_fps
        item['video_resolution'] = f'{v_width}x{v_height}'
        item['manufacturer'] = i.metadata.get('device', {}).get('manufacturer', '')
        item['model'] = i.metadata.get('device', {}).get('model', '')
        item['so_version'] = i.metadata.get('device', {}).get('androidVersion', '')
        
        items.append(item)

    return pd.DataFrame.from_dict(items)


def preprocess_sensors(data: dict, num_axes: int, datetime_format: str, start_time=None, end_time=None, debug=False):
  '''
  Converts the sensor data into appropriate data types and assigns time positions.
  '''
  series = {}
  ignored_lines = 0

  for row in data:
    ts = datetime.datetime.strptime(row['datetime_utc'], datetime_format)

    if start_time is not None and ts < start_time:
      ignored_lines += 1
      continue

    if end_time is not None and ts > end_time:
      ignored_lines += 1
      continue

    sensor_name = row['name']

    if sensor_name not in series:
      series[sensor_name] = []

    current_data = [ts]

    if num_axes >= 1:
      current_data.append(float(row['axis_x']))

    if num_axes >= 3:
      current_data.extend([
        float(row['axis_y']),
        float(row['axis_z']),
      ])

    if num_axes == 6:
      current_data.extend([
        float(row['delta_x']),
        float(row['delta_y']),
        float(row['delta_z']),
      ])

    series[sensor_name].append(current_data)

  if ignored_lines > 0 and debug:
    print(f'INFO. {ignored_lines} lines has been ignored.')

  for key, value in series.items():
      value.sort(key=lambda x: x[0])
      series[key] = to_dataframe(value, num_axes, datetime_format)

  return series


def preprocess_gps(data: dict):
	return np.array([[float(d['latitude']), float(d['longitude'])] for d in data])


def to_dataframe(data: dict, num_axes: int, datetime_format: str, create_time_column=True):
    '''
    Converts data into a Pandas.DataFrame and includes a column to represent the duration in seconds of the time series.
    Also returns the count of data points for each sensor axis.
    '''
    columns = ['Datetime UTC']

    if num_axes >= 1:
        columns.append('x')

    if num_axes >= 3:
        columns.extend(['y', 'z'])

    if num_axes == 6:
        columns.extend(['dx', 'dy', 'dz'])

    df = pd.DataFrame(data, columns=columns)

    if create_time_column:
        df['Datetime UTC'] = pd.to_datetime(df['Datetime UTC'], format=datetime_format)
        df['Time (s)'] = (df['Datetime UTC'] - df['Datetime UTC'].iloc[0]).dt.total_seconds()
        df = df.sort_values('Time (s)')

    return df


def resample_sensor_data(data: pd.DataFrame, target_fps=30):
    '''
    Converts the sensor data to match the FPS rate of the video.
    This is necessary because the frequency of sensor data varies.
    At times, there are 100 points in a single second, while at other times, there are only 50 points.
    This variation is controlled by the Android device, and it is crucial to ensure synchronization between the video and sensor data.
    '''
    resampled_data = pd.DataFrame()

    for second in range(int(data['Time (s)'].max()) + 1):
        interval_start = second
        interval_end = second + 1

        selected_lines = data[(data['Time (s)'] >= interval_start) & (data['Time (s)'] < interval_end)]

        indices = [int(x) for x in np.linspace(
            start=0,
            stop=len(selected_lines) - 1,
            num=target_fps)
        ]

        sample = selected_lines.iloc[indices]
        resampled_data = pd.concat([resampled_data, sample])

    resampled_data.reset_index(drop=True, inplace=True)

    return resampled_data


def inverse_geocode(latitude: float, longitude: float, key: str=None):
    data = {'country': 'Unknown', 'city': 'Unknown'}

    if not key:
        geo = reverse_geocode.search([[latitude, longitude]])
        if geo:
            first_geo = geo.pop()
            data['country'] = first_geo.get('country', 'Unknown')
            data['city'] = first_geo.get('city', 'Unknown')
    else:
        data = inverse_geocode_from_google(latitude, longitude, key)

    return data


def inverse_geocode_from_google(latitude: float, longitude: float, key: str):
    response = requests.get(f'https://maps.googleapis.com/maps/api/geocode/json?latlng={float(latitude)},{float(longitude)}&location_type=ROOFTOP&result_type=street_address&key={key}').json()
    data = {'country': 'Unknown', 'city': 'Unknown'}

    for r in response.get("results"):
        for i in r.get('address_components'):
            if 'country' in i.get('types'):
                data['country'] = i.get('long_name')

            if 'administrative_area_level_2' in i.get('types'):
                data['city'] = i.get('long_name')

    return data


def extract_sensor_snippet(data: pd.DataFrame, start_time, end_time, output_path):
    if data.empty:
        print('ERROR. The input DataFrame is empty.')
        return None
    
    required_columns = ['Time (s)',]
    for col in required_columns:
        if col not in data.columns:
            print(f"ERROR. Column '{col}' is not present in the DataFrame.")
            return None
    
    if not ((start_time >= data['Time (s)'].min()) and (end_time <= data['Time (s)'].max())):
        print('ERROR. The specified time range is outside the data range.')
        return None
    
    snippet = data[(data['Time (s)'] >= start_time) & (data['Time (s)'] <= end_time)]
    
    snippet.to_csv(output_path, sep=',', columns=data.columns, index=False)
    
    return snippet
